from optparse import OptionGroup, OptionParser, Values
from typing import Dict, List

from ..constants import IO_OPTIONS  # type: ignore
from ..log import *
from ..optparser._analysis import *
from ..version import __version__
from ._IO import *
from ._NN import *
from ._NT import *
from ._train import *


# ------------------------------------
# ONTraC functions
# ------------------------------------
def prepare_ontrac_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in this function first.
    :return: OptionParser object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC']  # type: ignore

    # usage and description
    usage = f'''USAGE: %prog <--NN-dir NN_DIR> <--GNN-dir GNN_DIR> <--NT-dir NT_DIR> <--meta-input META_INPUT> 
    [--n-cpu N_CPU] [--n-neighbors N_NEIGHBORS] [--n-local N_LOCAL] [--device DEVICE] [--epochs EPOCHS] [--patience PATIENCE]
    [--min-delta MIN_DELTA] [--min-epochs MIN_EPOCHS] [--batch-size BATCH_SIZE] [-s SEED] [--lr LR] [--hidden-feats HIDDEN_FEATS]
    [--n-gcn-layers N_GCN_LAYERS] [-k K] [--modularity-loss-weight MODULARITY_LOSS_WEIGHT] [--purity-loss-weight PURITY_LOSS_WEIGHT]
    [--regularization-loss-weight REGULARIZATION_LOSS_WEIGHT] [--beta BETA] [--trajectory-construct TRAJECTORY_CONSTRUCT]'''
    description = 'All steps of ONTraC including niche network construction, GNN, and niche construction.'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    # I/O options group
    add_IO_options_group(optparser=optparser, io_options=io_options)

    # Niche net construction options group
    add_niche_net_constr_options_group(optparser=optparser)

    # train and model options
    group_train: OptionGroup = add_train_options_group(optparser=optparser)
    add_GCN_options_group(group_train=group_train)
    add_GP_options_group(group_train=group_train)

    # Niche trajectory options group
    add_NT_options_group(optparser=optparser)

    return optparser


def _opt_ontrac_validate(options: Values,
                         io_options: Dict[str, List[str]],
                         optparser: Optional[OptionParser] = None) -> Values:
    """Validate options from a OptParser object.

    :param options: Options object.
    :param io_options: I/O options.
    :param optparser: OptionParser object.
    :return: Values object.
    """

    # IO
    validate_io_options(options=options, io_options=io_options, optparser=optparser)
    # niche network construction
    validate_niche_net_constr_options(options=options, optparser=optparser)
    # training
    validate_train_options(options=options, optparser=optparser)
    validate_GCN_options(options=options, optparser=optparser)
    validate_GP_options(options=options, optparser=optparser)
    # niche trajectory
    validate_NT_options(options=options, optparser=optparser)

    # print parameters to stdout
    info(message='------------------ RUN params memo ------------------ ')
    write_io_options_memo(options=options, io_options=io_options)
    write_niche_net_constr_memo(options=options)
    write_train_options_memo(options=options)
    write_GCN_options_memo(options=options)
    write_GP_options_memo(options=options)
    write_NT_options_memo(options=options)
    info(message='--------------- RUN params memo end ----------------- ')

    return options


def opt_ontrac_validate(optparser) -> Values:
    """Validate options from a OptParser object.

    :param optparser: OptionParser object.
    :return: Values object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC']  # type: ignore

    (options, args) = optparser.parse_args()

    options = _opt_ontrac_validate(options=options, io_options=io_options, optparser=optparser)

    return options


# ------------------------------------
# ONTraC_NN functions
# ------------------------------------
def prepare_nn_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    :return: OptionParser object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_NN']  # type: ignore

    # usage and description
    usage = f'''USAGE: %prog <--NN-dir NN_DIR> <--meta-input META_INPUT>
    [--n-cpu N_CPU] [--n-neighbors N_NEIGHBORS] [--n-local N_LOCAL]'''
    description = 'Create niche network and calculate features (normalized cell type composition). (Step 1 of ONTraC)'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    # I/O options group
    add_IO_options_group(optparser=optparser, io_options=io_options)

    # niche network construction options group
    add_niche_net_constr_options_group(optparser=optparser)

    return optparser


def _opt_nn_validate(options: Values,
                     io_options: Dict[str, List[str]],
                     optparser: Optional[OptionParser] = None) -> Values:
    """Validate options from a OptParser object.

    Parameters
    ----------
    options : Values
        Options object.
    io_options : Dict[str, List[str]]
        I/O options.
    optparser : Optional[OptionParser], optional
        OptionParser object, by default None

    Returns
    -------
    Values
        Values object.
    """

    # IO
    validate_io_options(options=options, io_options=io_options, optparser=optparser)
    # niche network construction
    validate_niche_net_constr_options(options=options, optparser=optparser)

    # print parameters to stdout
    info(message='------------------ RUN params memo ------------------ ')
    write_io_options_memo(options=options, io_options=io_options)
    write_niche_net_constr_memo(options=options)
    info(message='--------------- RUN params memo end ----------------- ')

    return options


def opt_nn_validate(optparser) -> Values:
    """Validate options from a OptParser object.

    :param optparser: OptionParser object.
    :return: Values object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_NN']  # type: ignore

    (options, args) = optparser.parse_args()

    options = _opt_nn_validate(options=options, io_options=io_options, optparser=optparser)

    return options


# ------------------------------------
# ONTraC_GNN functions
# ------------------------------------
def prepare_gnn_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    :return: OptionParser object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_GNN']  # type: ignore

    # usage and description
    usage = f'''USAGE: %prog <--NN-dir NN_DIR> <--GNN-dir GNN_DIR> [--device DEVICE]
    [--epochs EPOCHS] [--patience PATIENCE] [--min-delta MIN_DELTA] [--min-epochs MIN_EPOCHS] [--batch-size BATCH_SIZE] 
    [-s SEED] [--lr LR] [--hidden-feats HIDDEN_FEATS] [--n-gcn-layers N_GCN_LAYERS] [-k K]
    [--modularity-loss-weight MODULARITY_LOSS_WEIGHT] [--purity-loss-weight PURITY_LOSS_WEIGHT] 
    [--regularization-loss-weight REGULARIZATION_LOSS_WEIGHT] [--beta BETA]'''
    description = 'Graph Neural Network (GNN, GCN + GP). The core algorithm of ONTraC. (Step 2/3 of ONTraC)'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    # I/O options group
    add_IO_options_group(optparser=optparser, io_options=io_options)

    # train and model options
    group_train: OptionGroup = add_train_options_group(optparser=optparser)
    add_GCN_options_group(group_train=group_train)
    add_GP_options_group(group_train=group_train)

    return optparser


def _opt_gnn_validate(options: Values,
                      io_options: Dict[str, List[str]],
                      optparser: Optional[OptionParser] = None) -> Values:
    """Validate options from a OptParser object.

    Parameters
    ----------
    options : Values
        Options object.
    io_options : Dict[str, List[str]]
        I/O options.
    optparser : Optional[OptionParser], optional
        OptionParser object, by default None

    Returns
    -------
    Values
        Values object.
    """

    # IO
    validate_io_options(options=options, io_options=io_options, optparser=optparser)
    # training
    validate_train_options(options=options, optparser=optparser)
    validate_GCN_options(options=options, optparser=optparser)
    validate_GP_options(options=options, optparser=optparser)

    # print parameters to stdout
    info(message='------------------ RUN params memo ------------------ ')
    write_io_options_memo(options=options, io_options=io_options)
    write_train_options_memo(options=options)
    write_GCN_options_memo(options=options)
    write_GP_options_memo(options=options)
    info(message='--------------- RUN params memo end ----------------- ')

    return options


def opt_gnn_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    :param optparser: OptionParser object.
    :return: Values object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_GNN']  # type: ignore

    (options, args) = optparser.parse_args()

    options = _opt_gnn_validate(options=options, io_options=io_options, optparser=optparser)

    return options


# ------------------------------------
# ONTraC_NT functions
# ------------------------------------
def prepare_nt_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    :return: OptionParser object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_NT']  # type: ignore

    # usage and description
    usage = f'''USAGE: %prog <--NN-dir NN_DIR> <--GNN-dir GNN_DIR> <--NT-dir NT_DIR> 
            [--trajectory-construct TRAJECTORY_CONSTRUCT]'''
    description = 'ONTraC_NT: construct niche trajectory for niche cluster and project the NT score to each cell. (Step 4 of ONTraC)'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    add_IO_options_group(optparser=optparser, io_options=io_options)
    add_NT_options_group(optparser=optparser)

    return optparser


def _opt_nt_validate(options: Values,
                     io_options: Dict[str, List[str]],
                     optparser: Optional[OptionParser] = None) -> Values:
    """Validate options from a OptParser object.

    Parameters
    ----------
    options : Values
        Options object.
    io_options : Dict[str, List[str]]
        I/O options.
    optparser : Optional[OptionParser], optional
        OptionParser object, by default None

    Returns
    -------
    Values
        Values object.
    """

    # IO
    validate_io_options(options=options, io_options=io_options, optparser=optparser)
    # niche trajectory
    validate_NT_options(options=options, optparser=optparser)

    # print parameters to stdout
    info(message='------------------ RUN params memo ------------------ ')
    write_io_options_memo(options=options, io_options=io_options)
    write_NT_options_memo(options=options)
    info(message='--------------- RUN params memo end ----------------- ')

    return options


def opt_nt_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    :param optparser: OptionParser object.
    :return: Values object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_NT']  # type: ignore

    (options, args) = optparser.parse_args()

    options = _opt_nt_validate(options=options, io_options=io_options, optparser=optparser)

    return options


# ------------------------------------
# ONTraC_GT functions
# ------------------------------------
def prepare_gt_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    :return: OptionParser object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_GT']  # type: ignore

    usage = f'''USAGE: %prog <--NN-dir NN_DIR> <--GNN-dir GNN_DIR> <--NT-dir NT_DIR> [--device DEVICE]
    [--epochs EPOCHS] [--patience PATIENCE] [--min-delta MIN_DELTA] [--min-epochs MIN_EPOCHS] [--batch-size BATCH_SIZE] 
    [-s SEED] [--lr LR] [--hidden-feats HIDDEN_FEATS] [--n-gcn-layers N_GCN_LAYERS] [-k K]
    [--modularity-loss-weight MODULARITY_LOSS_WEIGHT] [--purity-loss-weight PURITY_LOSS_WEIGHT] 
    [--regularization-loss-weight REGULARIZATION_LOSS_WEIGHT] [--beta BETA] [--trajectory-construct TRAJECTORY_CONSTRUCT]'''
    description = 'ONTraC_GT: GNN and Niche Trajectory'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    # I/O options group
    add_IO_options_group(optparser=optparser, io_options=io_options)

    # train and model options
    group_train: OptionGroup = add_train_options_group(optparser=optparser)
    add_GCN_options_group(group_train=group_train)
    add_GP_options_group(group_train=group_train)

    # Niche trajectory
    add_NT_options_group(optparser=optparser)

    return optparser


def _opt_gt_validate(options: Values,
                     io_options: Dict[str, List[str]],
                     optparser: Optional[OptionParser] = None) -> Values:
    """Validate options from a OptParser object.

    Parameters
    ----------
    options : Values
        Options object.
    io_options : Dict[str, List[str]]
        I/O options.
    optparser : Optional[OptionParser], optional
        OptionParser object, by default None

    Returns
    -------
    Values
        Values object.
    """

    # IO
    validate_io_options(options=options, io_options=io_options, optparser=optparser)
    # training
    validate_train_options(options=options, optparser=optparser)
    validate_GCN_options(options=options, optparser=optparser)
    validate_GP_options(options=options, optparser=optparser)
    # niche trajectory
    validate_NT_options(options=options, optparser=optparser)

    # print parameters to stdout
    info(message='------------------ RUN params memo ------------------ ')
    write_io_options_memo(options, io_options)
    write_train_options_memo(options)
    write_GCN_options_memo(options)
    write_GP_options_memo(options)
    write_NT_options_memo(options)
    info(message='--------------- RUN params memo end ----------------- ')

    return options


def opt_gt_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    :param optparser: OptionParser object.
    :return: Values object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_GT']  # type: ignore

    (options, args) = optparser.parse_args()

    options = _opt_gt_validate(options=options, io_options=io_options, optparser=optparser)

    return options


# ------------------------------------
# ONTraC_analysis functions
# ------------------------------------
def prepare_analysis_optparser() -> OptionParser:
    """
    Prepare optparser object. New options will be added in thisfunction first.
    :return: OptionParser object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_analysis']  # type: ignore

    usage = f'''USAGE: %prog [--NN-dir NN_DIR] [--GNN-dir GNN_DIR] [--NT-dir NT_DIR] [-o OUTPUT]
    [--meta-input META_INPUT] [-l LOG] [-r REVERSE] [-s SAMPLE] [--scale-factor SCALE_FACTOR]
    [--suppress-cell-type-composition] [--suppress-niche-cluster-loadings] [--suppress-niche-trajectory]
    '''
    description = 'ONTraC_analysis: analysis of ONTraC results'

    # option processor
    optparser = OptionParser(version=f'%prog {__version__}', description=description, usage=usage, add_help_option=True)

    add_IO_options_group(optparser=optparser, io_options=io_options)
    add_visualization_group(optparser)
    add_suppress_group(optparser)

    return optparser


def _opt_analysis_validate(options: Values,
                           io_options: Dict[str, List[str]],
                           optparser: Optional[OptionParser] = None) -> Values:
    """Validate options from a OptParser object.

    Parameters
    ----------
    options : Values
        Options object.
    io_options : Dict[str, List[str]]
        I/O options.
    optparser : Optional[OptionParser], optional
        OptionParser object, by default None

    Returns
    -------
    Values
        Values object.
    """

    # IO
    validate_io_options(options=options, io_options=io_options, optparser=optparser)
    # visualization
    validate_visualization_options(options, optparser=optparser)
    validate_suppress_options(options, optparser=optparser)

    # print parameters to stdout
    info(message='------------------ RUN params memo ------------------ ')
    write_io_options_memo(options, io_options)
    write_visualization_options_memo(options)
    write_suppress_options_memo(options)
    info(message='--------------- RUN params memo end ----------------- ')

    return options


def opt_analysis_validate(optparser: OptionParser) -> Values:
    """Validate options from a OptParser object.

    :param optparser: OptionParser object.
    :return: Values object.
    """

    # args
    io_options: Dict[str, List[str]] = IO_OPTIONS['ONTraC_analysis']  # type: ignore

    (options, args) = optparser.parse_args()

    options = _opt_analysis_validate(options=options, io_options=io_options, optparser=optparser)

    return options


# ------------------------------------
# functions to be exported
# ------------------------------------
__all__ = [
    'prepare_ontrac_optparser', 'opt_ontrac_validate', 'prepare_nn_optparser', 'opt_nn_validate',
    'prepare_gnn_optparser', 'opt_gnn_validate', 'prepare_nt_optparser', 'opt_nt_validate', 'prepare_gt_optparser',
    'opt_gt_validate', 'prepare_analysis_optparser', 'opt_analysis_validate'
]



from optparse import Values

import pytest
import torch
from torch_geometric.loader import DenseDataLoader

from ONTraC.data import load_dataset
from ONTraC.GNN._GNN import SpatailOmicsDataset
from ONTraC.model import GNN
from ONTraC.train import GNNBatchTrain


@pytest.fixture
def options() -> Values:
    # Create an options object for testing
    _options = Values()
    _options.NN_dir = 'tests/_data/NN'
    _options.GNN_dir = 'tests/_data/GNN'
    _options.batch_size = 5
    _options.lr = 0.03
    _options.hidden_feats = 4
    _options.k = 6
    _options.modularity_loss_weight = 0.3
    _options.purity_loss_weight = 300
    _options.regularization_loss_weight = 0.1
    _options.beta = 0.03
    return _options


@pytest.fixture()
def dataset(options: Values) -> SpatailOmicsDataset:
    return load_dataset(NN_dir=options.NN_dir)


@pytest.fixture()
def sample_loader(options: Values, dataset: SpatailOmicsDataset) -> DenseDataLoader:
    batch_size = options.batch_size if options.batch_size > 0 else len(dataset)
    sample_loader = DenseDataLoader(dataset, batch_size=batch_size)
    return sample_loader


@pytest.fixture()
def nn_model(options: Values, dataset: SpatailOmicsDataset) -> torch.nn.Module:
    model = GNN(input_feats=dataset.num_features, hidden_feats=options.hidden_feats, k=options.k, exponent=options.beta)
    model.load_state_dict(torch.load(f'{options.GNN_dir}/epoch_0.pt', map_location=torch.device('cpu')))
    return model


def test_train(options: Values, sample_loader: DenseDataLoader, nn_model: torch.nn.Module) -> None:
    """
    Test the training process of GNN.
    :param options: options.
    :param sample_loader: DenseDataLoader, sample loader.
    :param nn_model: torch.nn.Module, GNN model.
    :return: None.
    """
    batch_train = GNNBatchTrain(model=nn_model, device=torch.device('cpu'), data_loader=sample_loader)
    optimizer = torch.optim.Adam(nn_model.parameters(), lr=options.lr)
    batch_train.set_train_args(optimizer=optimizer,
                               modularity_loss_weight=options.modularity_loss_weight,
                               purity_loss_weight=options.purity_loss_weight,
                               regularization_loss_weight=options.regularization_loss_weight,
                               beta=options.beta)
    batch_train.train_epoch(epoch=1)
    trained_params = torch.load(f'{options.GNN_dir}/epoch_1.pt', map_location=torch.device('cpu'))
    for k, v in nn_model.named_parameters():
        assert torch.allclose(v, trained_params[k],
                              rtol=0.05)  # there are some difference between linux and macOS (may be caused by chip?)



from pathlib import Path
from typing import Union

import pandas as pd

from ..log import warning


def read_meta_data(meta_data_file: Union[str, Path]) -> pd.DataFrame:
    """
    Read meta data file.
    :param meta_data_file: str or Path, meta data file path.
    :return: pd.DataFrame, meta data.
    """

    return pd.read_csv(meta_data_file, header=0, index_col=False, sep=',')


def valid_meta_data(meta_data_df: pd.DataFrame) -> pd.DataFrame:
    """
    Validate original data.
    :param meta_data_df: pd.DataFrame, original data.
    :return: pd.DataFrame, original data.

    1) check if Cell_ID, Sample, Cell_Type, x, and y columns in the original data
    2) make the Cell_Type column categorical
    3) return original data with Cell_ID/Spot_ID, Sample, Cell_Type, x, and y columns
    """

    # check if Cell_ID, Sample, Cell_Type, x, and y columns in the original data
    if 'Cell_ID' not in meta_data_df.columns and 'Spot_ID' not in meta_data_df.columns:
        raise ValueError('Cell_ID column is missing in the original data.')
    if 'Sample' not in meta_data_df.columns:
        raise ValueError('Sample column is missing in the original data.')
    if 'x' not in meta_data_df.columns:
        raise ValueError('x column is missing in the original data.')
    if 'y' not in meta_data_df.columns:
        raise ValueError('y column is missing in the original data.')

    id_name = 'Cell_ID'  # cell-level data

    # check if there any duplicated Cell_ID/Spot_ID in the original data
    if meta_data_df[id_name].duplicated().any():
        warning(
            f'There are duplicated {id_name} in the original data. Sample name will added to {id_name} to distinguish them.'
        )
        meta_data_df[id_name] = meta_data_df['Sample'] + '_' + meta_data_df[id_name]
    if meta_data_df[id_name].isnull().any():
        raise ValueError(f'Duplicated {id_name} within same sample found!')

    meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype(str).astype('category')
    # check if the cell type category number is less than 2
    if len(meta_data_df['Cell_Type'].cat.categories) < 2:
        raise ValueError('Cell_Type column found but less than 2 cell types.')
    meta_data_df = meta_data_df.dropna(subset=[id_name, 'Sample', 'Cell_Type', 'x', 'y'])

    # make the Sample column string
    meta_data_df['Sample'] = meta_data_df['Sample'].astype(str).astype('category')

    return meta_data_df


def save_cell_type_code(save_dir: Union[str, Path], cell_types: pd.Series) -> None:
    """
    Save mappings of the categorical data.
    :param save_dir: str or Path, save directory.
    :param cell_types: pd.Series, cell types.
    :return: None.
    """

    # check if the cell type is categorical
    if not pd.api.types.is_categorical_dtype(cell_types):  # type: ignore
        cell_types = cell_types.astype('category')

    # save mappings of the categorical data
    cell_type_code = pd.DataFrame(enumerate(cell_types.cat.categories), columns=['Code', 'Cell_Type'])
    cell_type_code.to_csv(f'{save_dir}/cell_type_code.csv', index=False)


def save_meta_data(save_dir: Union[str, Path], meta_data_df: pd.DataFrame) -> None:
    """
    Save meta data.
    :param save_dir: str or Path, save directory.
    :param meta_data_df: pd.DataFrame, meta data.
    :return: None.
    """

    meta_data_df.to_csv(f'{save_dir}/meta_data.csv.gz', index=False)


def load_meta_data(save_dir: Union[str, Path], meta_data_file: Union[str, Path]) -> pd.DataFrame:
    """
    Load original data.
    :param save_dir: str or Path, save directory.
    :return: pd.DataFrame, original data.
    """

    # read original data file
    meta_data_df = read_meta_data(meta_data_file=meta_data_file)

    meta_data_df = valid_meta_data(meta_data_df=meta_data_df)

    save_meta_data(save_dir=save_dir, meta_data_df=meta_data_df)

    return meta_data_df



from optparse import Values

from ..constants import IO_OPTIONS
from ..log import *
from ..optparser.command import (_opt_analysis_validate, _opt_gnn_validate, _opt_gt_validate, _opt_nn_validate,
                                 _opt_nt_validate, _opt_ontrac_validate)
from ..run.processes import gnn, niche_network_construct, niche_trajectory_construct


def options_valid(options: Values, process='ONTraC') -> Values:
    """
    Validate options
    :param options: options
    :param process: str, process name
    :return: options
    """

    # process should be a key in IO_OPTIONS
    if process not in IO_OPTIONS:
        raise ValueError(f'Invalid process name: {process}. Please choose one from {list(IO_OPTIONS.keys())}.')
    # get I/O options
    io_options = IO_OPTIONS[process]

    if process == 'ONTraC':
        options = _opt_ontrac_validate(options=options, io_options=io_options)
    elif process == 'ONTraC_NN':
        options = _opt_nn_validate(options=options, io_options=io_options)
    elif process == 'ONTraC_GNN':
        options = _opt_gnn_validate(options=options, io_options=io_options)
    elif process == 'ONTraC_NT':
        options = _opt_nt_validate(options=options, io_options=io_options)
    elif process == 'ONTraC_GT':
        options = _opt_gt_validate(options=options, io_options=io_options)
    elif process == 'ONTraC_analysis':
        options = _opt_analysis_validate(options=options, io_options=io_options)
    else:
        raise ValueError(f'Invalid process name: {process}. Please choose one from {list(IO_OPTIONS.keys())}.')

    return options


def run_ontrac(options: Values) -> None:
    """
    Run ONTraC
    :param options: options
    :return: None
    """

    # ----- options validation -----
    options = options_valid(options=options)

    # ----- Niche Network Construct -----
    niche_network_construct(options=options)

    # ----- GNN -----
    gnn(options=options)

    # ----- NT score -----
    niche_trajectory_construct(options=options)



#!/usr/bin/env python

import sys

from ..optparser import opt_nt_validate, prepare_nt_optparser
from ..run.processes import niche_trajectory_construct, load_parameters
from ..utils import write_version_info


# ------------------------------------
# Main Function
# ------------------------------------
def main() -> None:
    """
    The main function
    """

    # write version information
    write_version_info()

    # load parameters
    options = load_parameters(opt_validate_func=opt_nt_validate, prepare_optparser_func=prepare_nt_optparser)

    # ----- NT score -----
    niche_trajectory_construct(options=options)


# ------------------------------------
# Program running
# ------------------------------------
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.stderr.write("User interrupts me! ;-) See you ^.^!\n")
        sys.exit(0)



from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import pandas as pd

from ..log import info, warning
from .data import AnaData
from .utils import get_n_colors

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import pearsonr


def construct_meta_cell_along_trajectory(meta_data_df: pd.DataFrame,
                                         trajectory: str,
                                         n_cells: int = 10) -> pd.DataFrame:
    """
    Construct features for meta-cells by binning cells along the trajectory.

    Parameters
    ----------
    meta_data_df : pd.DataFrame
        DataFrame containing features for each cell. Rows are cells and columns are features. Features should be continuous values.
    trajectory : str
        Column name in feat_df that contains the trajectory values.
    n_cells : int
        Number of cells to bin in each meta-cell.
    
    Returns
    -------
    metacell_df : pd.DataFrame
        DataFrame containing features for meta-cells. Rows are meta-cells and columns are features.
    """

    # ensure trajectory in meta_data_df
    assert trajectory in meta_data_df.columns, f"Trajectory column {trajectory} not found in meta_data_df."

    # cell number should be at least 2*n_cells
    assert len(meta_data_df) >= 2 * n_cells, f"Number of cells in meta_data_df should be at least 2*n_cells."

    # sort meta_data_df by trajectory
    meta_data_df = meta_data_df.sort_values(by=trajectory)

    # bin cells into meta-cells
    rolled_mean = meta_data_df.rolling(window=n_cells).mean()
    mask = np.arange(len(rolled_mean)) % n_cells == 0
    mask = np.repeat(mask.reshape(-1, 1), rolled_mean.shape[1], axis=1)
    rolling_mean = rolled_mean.where(mask)
    rolling_mean = rolling_mean.dropna()

    return rolling_mean


def cal_features_correlation_along_trajectory(data_df: pd.DataFrame,
                                              trajectory: str,
                                              features: Optional[List[str]] = None,
                                              top_n: Optional[int] = None,
                                              rho_threshold: Optional[float] = None,
                                              p_val_threshold: Optional[float] = None) -> pd.DataFrame:
    """
    Calculate the correlation between features and a trajectory in a DataFrame.

    Parameters
    ----------
    data_df : pd.DataFrame
        DataFrame containing features for each (meta-)cell.
        Rows are meta-cells and columns are features and trajectory.
        All columns should be continuous values.
        All columns except the trajectory column should be features.
    trajectory : str
        Column name in data_df that contains the trajectory values.
    features : Optional[List[str]]
        Optional; if provided, only consider these features for correlation.
        If None, consider all features except the trajectory column.
    top_n : Optional[int]
        Optional; if provided, return only the top N features with highest absolute correlation to the trajectory.
        If None, return all features that meet the correlation and p-value thresholds.
    rho_threshold : Optional[float]
        Optional; minimum absolute correlation coefficient to consider a feature. If None, no threshold is applied.
    p_val_threshold : Optional[float]
        Optional; maximum p-value to consider a feature. If None, no threshold is applied.

    Returns
    -------
    pd.DataFrame
        DataFrame containing top correlated features.
    """

    if trajectory not in data_df.columns:
        raise ValueError(f"Trajectory column '{trajectory}' not found in metacell_data_df.")

    feature_list: List[str] = data_df.columns.difference(
        [trajectory]).tolist() if features is None else features
    correlations = []
    for feat in feature_list:
        rho, p_val = pearsonr(data_df[trajectory], data_df[feat])
        correlations.append((feat, rho, p_val))
    correlations_df = pd.DataFrame(correlations, columns=['Feature', 'PCC', 'P_Value'])
    correlations_df = correlations_df.dropna()
    correlations_df = correlations_df.sort_values(by='PCC', ascending=False)
    if rho_threshold is not None:
        correlations_df = correlations_df[abs(correlations_df['PCC']) >= rho_threshold]
    if p_val_threshold is not None:
        correlations_df = correlations_df[correlations_df['P_Value'] <= p_val_threshold]
    if top_n is not None:
        top_n_df = correlations_df.head(top_n)
        top_n_df = top_n_df[top_n_df['PCC']>0]
        bottom_n_df = correlations_df.tail(top_n)
        bottom_n_df = bottom_n_df[bottom_n_df['PCC']<0]
        correlations_df = pd.concat([correlations_df.head(top_n), correlations_df.tail(top_n)], ignore_index=True)
    return correlations_df.set_index('Feature')


def plot_scatter_feat_along_trajectory(data_df: pd.DataFrame,
                                       trajectory: str,
                                       feature: str,
                                       fit_reg: bool = True,
                                       annotate_pos: Union[str, None] = 'upper left',
                                       figszie: Tuple[int, int] = (5, 3),
                                       ylabel: Optional[str] = None,
                                       scatter_kws: Optional[Dict] = None,
                                       line_kws: Optional[Dict] = None,
                                       **kwargs) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot scatter plot of feature along trajectory.

    Parameters
    ----------
    data_df : pd.DataFrame
        DataFrame containing features for each cell. Rows are cells and columns are features. Features should be continuous values.
    trajectory : str
        Column name in feat_df that contains the trajectory values.
    feature : str
        Column name in feat_df that contains the feature values.
    fit_reg : bool
        Whether to plot regression line.
    annotate_pos : str
        Position to annotate PCC and p-value. Should be one of 'upper left', 'upper right', 'lower left', 'lower right', None.
        None means no annotation.
    figszie : Tuple[int, int]
        Figure size. Default is (5, 3).
    ylabel : str
        Label for y-axis.
    scatter_kws : dict
        Additional arguments for sns.regplot scatter_kws.
    line_kws : dict
        Additional arguments for sns.regplot line_kws.
    **kwargs
        Additional arguments for sns.regplot.

    Returns
    -------
    fig : plt.Figure
        Figure object.
    ax : plt.Axes
        Axes object.
    """

    # default parameters
    default_scatter_kws = {'edgecolor': None}
    default_line_kws = {'color': 'red'}
    default_kwargs = {'ci': None}

    # update default kws
    scatter_kws = {**default_scatter_kws, **(scatter_kws or {})}
    line_kws = {**default_line_kws, **(line_kws or {})}
    kwargs = {**default_kwargs, **kwargs}

    # ensure trajectory in data_df
    assert trajectory in data_df.columns, f"Trajectory column {trajectory} not found in data_df."

    # ensure feature in data_df
    assert feature in data_df.columns, f"Feature column {feature} not found in data_df."

    # sort data_df by trajectory
    data_df = data_df.sort_values(by=trajectory)

    # Visualization
    with sns.axes_style('white', rc={
            'xtick.bottom': True,
            'ytick.left': True
    }), sns.plotting_context('paper',
                             rc={
                                 'axes.titlesize': 8,
                                 'axes.labelsize': 8,
                                 'xtick.labelsize': 6,
                                 'ytick.labelsize': 6,
                                 'legend.fontsize': 6
                             }):
        fig, ax = plt.subplots(figsize=figszie)

        sns.regplot(data=data_df,
                    x=trajectory,
                    y=feature,
                    fit_reg=fit_reg,
                    scatter_kws=scatter_kws,
                    line_kws=line_kws,
                    **kwargs,
                    ax=ax)

        # descriptive annotation
        ax.set_title(feature)
        ax.set_xlabel(trajectory)
        ax.set_ylabel(ylabel if ylabel else feature)

        # correlation annotation
        rho, p_val = pearsonr(data_df[trajectory], data_df[feature])
        if annotate_pos == "upper left":
            ax.annotate(f"PCC = {rho:.2f}\nP = {p_val:.6f}",
                        xy=(0.02, 0.98),
                        xycoords="axes fraction",
                        ha="left",
                        va="top")
        elif annotate_pos == "upper right":
            ax.annotate(f"PCC = {rho:.2f}\nP = {p_val:.6f}",
                        xy=(0.98, 0.98),
                        xycoords="axes fraction",
                        ha="right",
                        va="top")
        elif annotate_pos == "lower left":
            ax.annotate(f"PCC = {rho:.2f}\nP = {p_val:.6f}",
                        xy=(0.02, 0.02),
                        xycoords="axes fraction",
                        ha="left",
                        va="bottom")
        elif annotate_pos == "lower right":
            ax.annotate(f"PCC = {rho:.2f}\nP = {p_val:.6f}",
                        xy=(0.98, 0.02),
                        xycoords="axes fraction",
                        ha="right",
                        va="bottom")
        elif annotate_pos is None:
            pass
        else:
            raise ValueError(f"annotate_pos should be one of 'upper left', 'upper right', 'lower left', 'lower right'")

    return fig, ax


def plot_cell_type_composition_along_trajectory(
        data_df: pd.DataFrame,
        trajectory: str,
        cell_types: Union[str, List[str]],
        agg_cell_num: int = 10,
        figsize: Tuple[int, int] = (6, 2),
        palette: Optional[Dict[str, str]] = None,
        output_file_path: Optional[Path] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot cell type composition along trajectory.

    Parameters
    ----------
    data_df : pd.DataFrame
        DataFrame containing cell type information for each cell. Rows are cells and columns are cell types and trajectory.
    trajectory : str
        Column name in data_df that contains the trajectory values.
    cell_types : str or list of str
        Column name(s) in data_df that contains the cell type information.
    agg_cell_num : int
        Number of cells to aggregate in each bin along the trajectory. Default is 10. 1 means no aggregation.
    figsize : Optional[Tuple[int, int]]
        Figure size. Default is (6, 2).
    palette : Optional[Dict[str, str]]
        Color palette for cell types. If None, use default color palette. Keys are cell types and values are colors.
    output_file_path : Optional[Path]
        If provided, save the figure to this path. If None, do not save the figure and return the figure, and axes objects.
    
    Returns
    -------
    fig : plt.Figure
        Figure object.
    ax : plt.Axes
        Axes object.
    """

    # ensure trajectory in data_df
    assert trajectory in data_df.columns, f"Trajectory column {trajectory} not found in data_df."

    # ensure cell_types in data_df
    if isinstance(cell_types, str):
        cell_types = [cell_types]
    for cell_type in cell_types:
        assert cell_type in data_df.columns, f"Cell type column {cell_type} not found in data_df."

    # if palette is provided, all cell types should be in palette
    if palette is not None:
        for cell_type in cell_types:
            assert cell_type in palette, f"Cell type {cell_type} not found in palette."
    else:
        palette = {cell_type: color for cell_type, color in zip(cell_types, get_n_colors(len(cell_types)))}

    # sort data_df by trajectory
    data_df = data_df.sort_values(by=trajectory)

    # bin cells into meta-cells
    if agg_cell_num > 1:
        rolling_mean = construct_meta_cell_along_trajectory(meta_data_df=data_df.loc[:, cell_types + [trajectory]],
                                                            trajectory=trajectory,
                                                            n_cells=agg_cell_num)
    else:
        rolling_mean = data_df

    # plot cell type composition along trajectory
    with sns.axes_style('white', rc={
            'xtick.bottom': True,
            'ytick.left': True
    }), sns.plotting_context('paper',
                             rc={
                                 'axes.titlesize': 8,
                                 'axes.labelsize': 8,
                                 'xtick.labelsize': 6,
                                 'ytick.labelsize': 6,
                                 'legend.fontsize': 6
                             }):
        fig, ax = plt.subplots(figsize=figsize)

        for cell_type in cell_types:
            ax.plot(rolling_mean[trajectory], rolling_mean[cell_type], label=cell_type, color=palette[cell_type])

        ax.set_ylim(0, 1)
        ax.legend(title='Cell Types', bbox_to_anchor=(1.05, 1), loc='upper left', ncol=2)
        ax.set_xlabel(trajectory)
        ax.set_ylabel('Normalized Cell Type Composition')
        ax.set_title('Cell Type Composition Along Niche Trajectory')
        fig.tight_layout()

    if output_file_path is not None:
        output_file_path.parent.mkdir(parents=True, exist_ok=True)
        fig.savefig(output_file_path, bbox_inches='tight', dpi=300)
        info(f"Saved cell type composition plot to {output_file_path}")
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_cell_type_composition_along_trajectory_from_anadata(
        ana_data: AnaData,
        cell_types: Optional[Union[str, List[str]]] = None,
        agg_cell_num: int = 10,
        figsize: Tuple[int, int] = (6, 2),
        palette: Optional[Dict[str, str]] = None,
        output_file_path: Optional[Union[Path, str]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot cell type composition (niche features) along trajectory from AnaData object.

    Parameters
    ----------
    ana_data : AnaData
        AnaData object.
    cell_types : str or list of str
        Column name(s) in AnaData.meta_data_df that contains the cell type information.
        Default is None, which means all cell types in AnaData.cell_type_codes will be used.
    agg_cell_num : int
        Number of cells to aggregate in each bin along the trajectory. Default is 10. 1 means no aggregation.
    figsize : Tuple[int, int]
        Figure size. Default is (6, 2).
    palette : Optional[Dict[str, str]]
        Color palette for cell types. If None, use default color palette. Keys are cell types and values are colors.
    output_file_path : Optional[Union[Path, str]]
        Path to save the figure. If None, the default path
        {ana_data.options.output}/lineplot_raw_cell_type_composition_along_trajectory.pdf is used. 
        If ana_data.options.output is also None, the figure will not be saved and the function 
        will return the figure and axes objects instead.
    
    Returns
    -------
    fig : plt.Figure
        Figure object.
    ax : plt.Axes
        Axes object.
    """

    # ensure ana_data has trajectory and cell types
    if ana_data.NT_score is None:
        warning("NT score is not set in AnaData object. Skipping cell type composition along trajectory plot.")
    if ana_data.cell_type_codes is None:
        warning("Cell type codes are not set in AnaData object. Skipping cell type composition along trajectory plot.")
    if ana_data.cell_type_composition is None:
        warning(
            "Cell type composition is not set in AnaData object. Skipping cell type composition along trajectory plot.")

    # data_df
    data_df = ana_data.meta_data_df.copy()
    data_df = data_df.join(1 - ana_data.NT_score['Cell_NTScore'] if hasattr(ana_data.options, 'reverse')
                           and ana_data.options.reverse else ana_data.NT_score['Cell_NTScore'])  # type: ignore
    data_df = data_df.join(ana_data.cell_type_composition)

    # cell types
    if cell_types is None:
        cell_types = ana_data.cell_type_codes['Cell_Type'].values.tolist()
    elif isinstance(cell_types, str):
        cell_types = [cell_types]

    # output file path
    if output_file_path is None:
        if not hasattr(ana_data.options, 'output') or ana_data.options.output is None:
            output_file_path = None
        else:
            output_file_path = Path(ana_data.options.output) / 'lineplot_raw_cell_type_composition_along_trajectory.pdf'
    else:
        output_file_path = Path(output_file_path)
        if not output_file_path.suffix:
            output_file_path = output_file_path.with_suffix('.pdf')

    return plot_cell_type_composition_along_trajectory(
        data_df=data_df,
        trajectory='Cell_NTScore',
        cell_types=cell_types,  # type: ignore
        agg_cell_num=agg_cell_num,
        figsize=figsize,
        palette=palette,
        output_file_path=output_file_path)



import os
from optparse import Values
from typing import Dict, Optional

import numpy as np
import pandas as pd
from pandas import DataFrame

from ..log import warning
from ..utils import get_meta_data_file, get_rel_params, read_yaml_file


# ----------------------------
# Misc Functions
# ----------------------------
def load_loss_record_data(options) -> Optional[Dict]:
    """
    Parse the log file and save the training loss to a csv file.
    Args:
        options: Values, the options from optparse
    Returns:
        Dict: {'loss_df': pd.DataFrame, 'loss_dict': Dict} or None
        loss_df is the training loss dataframe
        loss_dict is the final loss dictionary

    loss record example:
        16:31:47 --- INFO: epoch: 1, batch: 1, loss: 8.554190635681152, spectral_loss: -6.64739809508319e-06, cluster_loss: 0.06959901750087738, feat_similarity_loss: 8.484598159790039
    """
    loss, loss_name = [], []  # type: ignore
    if options.log is None:
        return None
    with open(options.log, 'r') as fhd:
        for line in fhd:  # type: ignore
            # train loss record
            if 'INFO' in line and 'epoch' in line and 'loss' in line:
                loss_ = []
                loss_name = []
                line: list[str] = line.strip().split()  # type: ignore
                init_index = line.index('loss:')  # get the first loss index
                # get loss name & losss
                for i in range(init_index, len(line), 2):  # loss name
                    loss_name.append(line[i].strip(':'))
                for i in range(init_index + 1, len(line), 2):  # loss value
                    loss_.append(float(line[i].strip(',')))
                # insert epoch and batch information
                loss_.insert(0, int(line[init_index - 1].strip(',')))  # batch index
                loss_.insert(0, int(line[init_index - 3].strip(',')))  # epoch index
                loss.append(loss_)
            # eval loss record
            elif 'INFO' in line and ('Evaluation loss' in line or 'Evaluate loss' in line):
                line = line.strip().split(', ', 1)  # type: ignore
                final_loss_dict: Dict[str, float] = eval(line[1])

    # loss name edit
    loss_name[0] = 'total_loss'
    loss_name = ['Epoch', 'Batch'] + loss_name[:]

    # save loss to csv
    loss_df = pd.DataFrame(loss, columns=loss_name)
    if options.output is not None:
        loss_df.to_csv(f'{options.output}/train_loss.csv', index=False)

    # calculate epoch loss
    epoch_loss_columns = [col for col in loss_df.columns if 'loss' in col]
    epoch_loss_columns.insert(0, 'Epoch')
    epoch_loss_df = loss_df[epoch_loss_columns].groupby('Epoch').mean()
    if options.output is not None:
        epoch_loss_df.to_csv(f'{options.output}/train_loss_epoch.csv')

    return {'loss_df': epoch_loss_df, 'loss_dict': final_loss_dict}


def load_niche_cluster_connectivity(options: Values) -> Optional[np.ndarray]:
    """
    Load the niche cluster connectivity from the output of GNN.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[np.ndarray]: the niche cluster connectivity
    """

    if options.GNN_dir is None:
        return None

    niche_cluster_conn_file = f'{options.GNN_dir}/consolidate_out_adj.csv.gz'
    if not os.path.isfile(niche_cluster_conn_file):
        niche_cluster_conn_file = f'{options.GNN_dir}/consolidate_out_adj.csv'
    if not os.path.isfile(niche_cluster_conn_file):  # skip if file not exist
        warning(f"Cannot find niche cluster connectivity file: {niche_cluster_conn_file}.")
        return None

    return np.loadtxt(f'{niche_cluster_conn_file}', delimiter=',')


def load_niche_cluster_score(options: Values) -> Optional[np.ndarray]:
    """
    Load the niche cluster score from the output of NT score.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[np.ndarray]: the niche cluster score
    """

    if options.NT_dir is None:
        return None

    niche_cluster_score_file = f'{options.NT_dir}/niche_cluster_score.csv.gz'
    if not os.path.isfile(niche_cluster_score_file):
        niche_cluster_score_file = f'{options.NT_dir}/niche_cluster_score.csv'
    if not os.path.isfile(niche_cluster_score_file):  # skip if file not exist
        warning(f"Cannot find niche cluster score file: {niche_cluster_score_file}.")
        return None

    return np.loadtxt(f'{niche_cluster_score_file}', delimiter=',')


def load_niche_level_niche_cluster_assign(options: Values) -> Optional[pd.DataFrame]:
    """
    Load the niche cluster assignment for each niche level.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[pd.DataFrame]: the niche cluster assignment for each niche level
    """

    if options.GNN_dir is None:
        return None

    niche_level_niche_cluster_assign_file = f'{options.GNN_dir}/niche_level_niche_cluster.csv.gz'
    if not os.path.isfile(niche_level_niche_cluster_assign_file):
        niche_level_niche_cluster_assign_file = f'{options.GNN_dir}/niche_level_niche_cluster.csv'
    if not os.path.isfile(niche_level_niche_cluster_assign_file):  # skip if file not exist
        warning(f"Cannot find niche level niche cluster assign file: {niche_level_niche_cluster_assign_file}.")
        return None

    return pd.read_csv(niche_level_niche_cluster_assign_file, index_col=0)


def load_cell_level_niche_cluster_assign(options: Values) -> Optional[pd.DataFrame]:
    """
    Load the niche cluster assignment for each cell level.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[pd.DataFrame]: the niche cluster assignment for each cell level
    """

    if options.GNN_dir is None:
        return None

    cell_level_niche_cluster_assign_file = f'{options.GNN_dir}/cell_level_niche_cluster.csv.gz'
    if not os.path.isfile(cell_level_niche_cluster_assign_file):
        cell_level_niche_cluster_assign_file = f'{options.GNN_dir}/cell_level_niche_cluster.csv'
    if not os.path.isfile(cell_level_niche_cluster_assign_file):  # skip if file not exist
        warning(f"Cannot find cell level niche cluster assign file: {cell_level_niche_cluster_assign_file}.")
        return None

    return pd.read_csv(cell_level_niche_cluster_assign_file, index_col=0)


def load_niche_level_max_niche_cluster(options: Values) -> Optional[pd.DataFrame]:
    """
    Load the max niche cluster assignment for each niche level.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[pd.DataFrame]: the max niche cluster assignment for each niche level
    """

    if options.GNN_dir is None:
        return None

    niche_level_max_niche_cluster_file = f'{options.GNN_dir}/niche_level_max_niche_cluster.csv.gz'
    if not os.path.isfile(niche_level_max_niche_cluster_file):
        niche_level_max_niche_cluster_file = f'{options.GNN_dir}/niche_level_max_niche_cluster.csv'
    if not os.path.isfile(niche_level_max_niche_cluster_file):  # skip if file not exist
        warning(f"Cannot find niche level max niche cluster file: {niche_level_max_niche_cluster_file}.")
        return None

    return pd.read_csv(niche_level_max_niche_cluster_file, index_col=0)


def load_cell_level_max_niche_cluster(options: Values) -> Optional[pd.DataFrame]:
    """
    Load the max niche cluster assignment for each cell level.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[pd.DataFrame]: the max niche cluster assignment for each cell level
    """

    if options.GNN_dir is None:
        return None

    cell_level_max_niche_cluster_file = f'{options.GNN_dir}/cell_level_max_niche_cluster.csv.gz'
    if not os.path.isfile(cell_level_max_niche_cluster_file):
        cell_level_max_niche_cluster_file = f'{options.GNN_dir}/cell_level_max_niche_cluster.csv'
    if not os.path.isfile(cell_level_max_niche_cluster_file):  # skip if file not exist
        warning(f"Cannot find cell level max niche cluster file: {cell_level_max_niche_cluster_file}.")
        return None

    return pd.read_csv(cell_level_max_niche_cluster_file, index_col=0)


def load_niche_hidden_features(options: Values) -> Optional[np.ndarray]:
    """
    Load the niche hidden features.
    Args:
        options: Values, the options from optparse
    Returns:
        Optional[np.ndarray]: the niche hidden features
    """

    if options.NT_dir is None:
        return None

    niche_hidden_features_file = f'{options.NT_dir}/niche_hidden_features.csv.gz'
    if not os.path.isfile(niche_hidden_features_file):
        niche_hidden_features_file = f'{options.NT_dir}/niche_hidden_features.csv'
    if not os.path.isfile(niche_hidden_features_file):  # skip if file not exist
        warning(f"Cannot find niche hidden features file: {niche_hidden_features_file}.")
        return None

    return np.loadtxt(f'{niche_hidden_features_file}', delimiter=',')


# ----------------------------
# Classes
# ----------------------------


class AnaData:
    """
    Class to store the data for analysis
    This class have the following attributes:
    - options: Values, the options from optparse
    - rel_params: Dict, the relative paths for params
    - meta_data_df: pd.DataFrame, the original Cell ID and Cell Type
    - train_loss: Dict, the training loss
    - cell_type_codes: pd.DataFrame, the cell type codes
    - cell_type_composition: pd.DataFrame, the cell type composition
    - NT_score: pd.DataFrame, the NT score
    - niche_cluster_connectivity: np.ndarray, the niche cluster connectivity
    - niche_cluster_score: np.ndarray, the niche cluster score
    - niche_level_niche_cluster_assign: pd.DataFrame, the niche cluster assignment for each niche level
    - cell_level_niche_cluster_assign: pd.DataFrame, the niche cluster assignment for each cell level
    - niche_level_max_niche_cluster: pd.DataFrame, the max niche cluster assignment for each niche level
    - cell_level_max_niche_cluster: pd.DataFrame, the max niche cluster assignment for each cell level
    - niche_hidden_features: np.ndarray, the hidden features for each niche level
    """

    def __init__(self, options: Values) -> None:
        """
        Initialize the class with the options"""

        # save options
        self.options = options

        # meta_data_df
        if hasattr(self.options, 'NN_dir'):

            # get real path
            params = read_yaml_file(f'{options.NN_dir}/samples.yaml')
            self.rel_params = get_rel_params(options.NN_dir, params)
            # save the original Cell_ID
            self.meta_data_df = pd.read_csv(get_meta_data_file(options.NN_dir))
            self.meta_data_df = self.meta_data_df.set_index('Cell_ID')
        else:  # not NN_dir, only support for visualization of meta_input
            self.meta_data_df = pd.read_csv(self.options.meta_input)
            self.meta_data_df = self.meta_data_df.set_index('Cell_ID')

        # make the Cell_Type column categorical
        # the order of categories is the same as the order of appearance in the cell_type_codes
        self.meta_data_df['Cell_Type'] = self.meta_data_df['Cell_Type'].astype('category')

        # make the Sample column categorical
        self.meta_data_df['Sample'] = self.meta_data_df['Sample'].astype('str').astype('category')

    @property
    def train_loss(self):
        if not hasattr(self, '_train_loss') or self._train_loss is None:
            self._train_loss = load_loss_record_data(self.options)
        return self._train_loss

    @property
    def cell_type_codes(self) -> pd.DataFrame:
        if not hasattr(self, '_cell_type_codes') or self._cell_type_codes is None:  # type: ignore
            self._cell_type_codes = pd.read_csv(f'{self.options.NN_dir}/cell_type_code.csv', index_col=0)
            # order the cell type in meta_data_df
            self.meta_data_df['Cell_Type'] = pd.Categorical(self.meta_data_df['Cell_Type'],
                                                            categories=self._cell_type_codes['Cell_Type'].tolist())
        return self._cell_type_codes

    def _load_cell_type_composition(self) -> None:
        data_df = pd.DataFrame()
        for sample in self.rel_params['Data']:
            feature_file = sample['Features']
            if not os.path.isfile(feature_file):
                raise FileNotFoundError(f"Cannot find cell type composition file: {feature_file}.")
            cell_type_composition_df = pd.read_csv(feature_file, header=None)
            cell_type_composition_df.index = self.meta_data_df[self.meta_data_df['Sample'] == sample['Name']].index
            cell_type_composition_df.columns = self.cell_type_codes.loc[np.arange(cell_type_composition_df.shape[1]),
                                                                        'Cell_Type'].tolist()  # type: ignore
            data_df = pd.concat([data_df, cell_type_composition_df])
        self._cell_type_composition = data_df.loc[self.meta_data_df.index]

    @property
    def cell_type_composition(self) -> DataFrame:
        if not hasattr(self, '_cell_type_composition') or self._cell_type_composition is None:
            self._load_cell_type_composition()
        return self._cell_type_composition

    def _load_NT_score(self) -> Optional[DataFrame]:
        if self.options.NT_dir is None:
            return None
        if not os.path.isfile(f'{self.options.NT_dir}/NTScore.csv.gz'):
            warning(f"Cannot find NT score file: {self.options.NT_dir}/NTScore.csv.gz.")
            return None
        NTScore_df = pd.read_csv(f'{self.options.NT_dir}/NTScore.csv.gz', index_col=0)
        return NTScore_df.loc[self.meta_data_df.index]

    @property
    def NT_score(self) -> Optional[DataFrame]:
        if not hasattr(self, '_NT_score') or self._NT_score is None:  # type: ignore
            self._NT_score = self._load_NT_score()
        return self._NT_score

    @property
    def niche_cluster_connectivity(self) -> Optional[np.ndarray]:
        if not hasattr(self, '_niche_cluster_connectivity') or self._niche_cluster_connectivity is None:  # type: ignore
            self._niche_cluster_connectivity = load_niche_cluster_connectivity(self.options)
        return self._niche_cluster_connectivity

    @property
    def niche_cluster_score(self) -> Optional[np.ndarray]:
        if not hasattr(self, '_niche_cluster_score') or self._niche_cluster_score is None:  # type: ignore
            self._niche_cluster_score = load_niche_cluster_score(self.options)
        return self._niche_cluster_score

    @property
    def niche_level_niche_cluster_assign(self) -> Optional[pd.DataFrame]:
        if not hasattr(
                self,
                '_niche_level_niche_cluster_assign') or self._niche_level_niche_cluster_assign is None:  # type: ignore
            self._niche_level_niche_cluster_assign = load_niche_level_niche_cluster_assign(self.options)
            if self._niche_level_niche_cluster_assign is None:
                return None
            try:
                self._niche_level_niche_cluster_assign = self._niche_level_niche_cluster_assign.loc[
                    self.meta_data_df.index]
            except:
                pass
        return self._niche_level_niche_cluster_assign

    @property
    def cell_level_niche_cluster_assign(self) -> Optional[pd.DataFrame]:
        if not hasattr(
                self,
                '_cell_level_niche_cluster_assign') or self._cell_level_niche_cluster_assign is None:  # type: ignore
            self._cell_level_niche_cluster_assign = load_cell_level_niche_cluster_assign(self.options)
            if self._cell_level_niche_cluster_assign is None:
                return None
            try:
                self._cell_level_niche_cluster_assign = self._cell_level_niche_cluster_assign.loc[
                    self.meta_data_df.index]
            except:
                pass
        return self._cell_level_niche_cluster_assign

    @property
    def niche_level_max_niche_cluster(self) -> Optional[pd.DataFrame]:
        if not hasattr(self,
                       '_niche_level_max_niche_cluster') or self._niche_level_max_niche_cluster is None:  # type: ignore
            self._niche_level_max_niche_cluster = load_niche_level_max_niche_cluster(self.options)
            if self._niche_level_max_niche_cluster is None:
                return None
            try:
                self._niche_level_max_niche_cluster = self._niche_level_max_niche_cluster.loc[self.meta_data_df.index]
            except:
                pass
        return self._niche_level_max_niche_cluster

    @property
    def cell_level_max_niche_cluster(self) -> Optional[pd.DataFrame]:
        if not hasattr(self,
                       '_cell_level_max_niche_cluster') or self._cell_level_max_niche_cluster is None:  # type: ignore
            self._cell_level_max_niche_cluster = load_cell_level_max_niche_cluster(self.options)
            if self._cell_level_max_niche_cluster is None:
                return None
            try:
                self._cell_level_max_niche_cluster = self._cell_level_max_niche_cluster.loc[self.meta_data_df.index]
            except:
                pass
        return self._cell_level_max_niche_cluster

    @property
    def niche_hidden_features(self) -> Optional[np.ndarray]:
        if not hasattr(self, '_niche_hidden_features') or self._niche_hidden_features is None:  # type: ignore
            self._niche_hidden_features = load_niche_hidden_features(self.options)
            if self._niche_hidden_features is None:
                return None
            try:
                self._niche_hidden_features = self._niche_hidden_features[self.meta_data_df.index]
            except:
                pass
        return self._niche_hidden_features



from .version import __version__

citation = "Wang, W., Zheng, S., Shin, C. S. & Yuan, G. C. Characterizing Spatially Continuous Variations in Tissue Microenvironment through Niche Trajectory Analysis. https://www.biorxiv.org/content/10.1101/2024.04.23.590827v1. bioRxiv, 2024."


from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from scipy.sparse import load_npz

from ..log import info
from .algorithm import brute_force


def get_niche_trajectory_path(trajectory_construct_method: str, niche_adj_matrix: ndarray) -> List[int]:
    """
    Get niche trajectory path
    :param trajectory_construct_method: str, the method to construct trajectory
    :param adj_matrix: non-negative ndarray, adjacency matrix of the graph
    :return: List[int], the niche trajectory
    """

    niche_adj_matrix = (niche_adj_matrix + niche_adj_matrix.T) / 2

    if trajectory_construct_method == 'BF':
        info('Finding niche trajectory with maximum connectivity using Brute Force.')

        niche_trajectory_path = brute_force(niche_adj_matrix)

    return niche_trajectory_path


def trajectory_path_to_NC_score(niche_trajectory_path: List[int]) -> ndarray:
    """
    Convert niche cluster trajectory path to NTScore
    :param niche_trajectory_path: List[int], the niche trajectory path
    :return: ndarray, the NTScore
    """

    info('Calculating NTScore for each niche cluster based on the trajectory path.')

    niche_NT_score = np.zeros(len(niche_trajectory_path))
    values = np.linspace(0, 1, len(niche_trajectory_path))

    for i, index in enumerate(niche_trajectory_path):
        # debug(f'i: {i}, index: {index}')
        niche_NT_score[index] = values[i]
    return niche_NT_score


def get_niche_NTScore(trajectory_construct_method: str, niche_level_niche_cluster_assign_df: DataFrame,
                      niche_adj_matrix: ndarray) -> Tuple[ndarray, DataFrame]:
    """
    Get niche-level niche trajectory and cell-level niche trajectory
    :param trajectory_construct_method: str, the method to construct trajectory
    :param niche_level_niche_cluster_assign_df: DataFrame, the niche-level niche cluster assignment. #niche x #niche_cluster
    :param adj_matrix: ndarray, the adjacency matrix of the graph
    :return: Tuple[ndarray, DataFrame], the niche-level niche trajectory and cell-level niche trajectory
    """

    info('Calculating NTScore for each niche.')

    niche_trajectory_path = get_niche_trajectory_path(trajectory_construct_method=trajectory_construct_method,
                                                      niche_adj_matrix=niche_adj_matrix)

    niche_cluster_score = trajectory_path_to_NC_score(niche_trajectory_path)
    niche_level_NTScore_df = pd.DataFrame(niche_level_niche_cluster_assign_df.values @ niche_cluster_score,
                                          index=niche_level_niche_cluster_assign_df.index,
                                          columns=['Niche_NTScore'])
    return niche_cluster_score, niche_level_NTScore_df


def niche_to_cell_NTScore(meta_data_df: DataFrame, niche_level_NTScore_df: DataFrame, rel_params: Dict) -> DataFrame:
    """
    get cell-level NTScore
    :param meta_data_df: DataFrame, the meta data
    :param niche_level_NTScore_df: DataFrame, the niche-level NTScore
    :param rel_params: Dict, relative paths
    :return: DataFrame, cell-level NTScore
    """

    info('Projecting NTScore from niche-level to cell-level.')

    # prepare
    id_name: str = meta_data_df.columns[0]
    samples = meta_data_df['Sample'].cat.categories
    sample_files_by_name_dict = {
        sample_files_dict['Name']: sample_files_dict
        for sample_files_dict in rel_params['Data']
    }
    sample_cell_level_NTScore_list = []

    for sample in samples:
        sample_niche_level_NTScore_df = niche_level_NTScore_df.loc[meta_data_df[meta_data_df['Sample'] == sample][id_name].values]
        niche_weight_matrix = load_npz(sample_files_by_name_dict[sample]['NicheWeightMatrix'])
        if niche_weight_matrix.shape[0] != sample_niche_level_NTScore_df.shape[0]:
            raise ValueError(f'Inconsistent number of niches in {sample} sample. '
                             f'Please check the niche weight matrix and the niche-level NTScore.')
        if niche_weight_matrix.shape[1] != sample_niche_level_NTScore_df.shape[0]:
            raise ValueError(f'Inconsistent number of cells in {sample} sample. '
                             f'Please check the niche weight matrix and the niche-level NTScore.')
        niche_to_cell_matrix = (niche_weight_matrix / niche_weight_matrix.sum(axis=0)
                                ).T  # normalize by the all niches associated with each cell, N (#cell) x N (#niche)

        # cell-level NTScore
        niche_level_NTScore_ = sample_niche_level_NTScore_df.values.reshape(-1, 1)  # N x 1
        cell_level_NTScore_ = niche_to_cell_matrix @ niche_level_NTScore_
        sample_cell_level_NTScore_list.append(
            pd.DataFrame(cell_level_NTScore_.reshape(-1),
                         index=sample_niche_level_NTScore_df.index,
                         columns=['Cell_NTScore']))

    cell_level_NTScore_df = pd.concat(sample_cell_level_NTScore_list).loc[meta_data_df[id_name]]

    return cell_level_NTScore_df


def NTScore_table(save_dir: Union[str, Path], meta_data_df: DataFrame, niche_level_NTScore_df: DataFrame,
                  cell_level_NTScore_df: DataFrame, rel_params: Dict) -> None:
    """
    Generate NTScore table and save it
    :param save_dir: Union[str, Path], the directory to save NTScore table
    :param meta_data_df: DataFrame, the meta data
    :param niche_level_NTScore_df: DataFrame, the niche-level NTScore
    :param cell_level_NTScore_df: DataFrame, the cell-level NTScore
    :param rel_params: Dict, relative paths
    :return: None
    """

    info('Output NTScore tables.')

    # prepare
    id_name: str = meta_data_df.columns[0]
    samples = meta_data_df['Sample'].cat.categories
    sample_files_by_name_dict = {
        sample_files_dict['Name']: sample_files_dict
        for sample_files_dict in rel_params['Data']
    }
    NTScore_table = pd.DataFrame()

    for sample in samples:
        coordinates_df = pd.read_csv(sample_files_by_name_dict[sample]['Coordinates'], index_col=0)
        sample_niche_level_NTScore_df = niche_level_NTScore_df.loc[meta_data_df[meta_data_df['Sample'] == sample][id_name].values]
        sample_cell_level_NTScore_df = cell_level_NTScore_df.loc[meta_data_df[meta_data_df['Sample'] == sample][id_name].values]
        coordinates_df = coordinates_df.join(sample_niche_level_NTScore_df).join(sample_cell_level_NTScore_df)
        coordinates_df.to_csv(f'{save_dir}/{sample}_NTScore.csv.gz')
        NTScore_table = pd.concat([NTScore_table, coordinates_df])

    NTScore_table.loc[meta_data_df[id_name]].to_csv(f'{save_dir}/NTScore.csv.gz')



import itertools
from typing import List

import numpy as np


def brute_force(conn_matrix: np.ndarray) -> List[int]:
    """
    Brute force method to find the optimal path with the highest connectivity.
    """
    max_connectivity = float('-inf')
    optimal_path = []
    for path in itertools.permutations(range(len(conn_matrix))):
        connectivity = 0
        for i in range(len(path) - 1):
            connectivity += conn_matrix[path[i], path[i + 1]]
        if connectivity > max_connectivity:
            max_connectivity = connectivity
            optimal_path = list(path)
    return optimal_path



from math import inf
import sys
from optparse import OptionGroup, OptionParser, Values
from random import randint
from typing import Optional

import torch

from ..log import *


def add_train_options_group(optparser: OptionParser) -> OptionGroup:
    """
    Add train options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """
    # overall train options group
    group_train = OptionGroup(optparser, "Options for GNN training")
    optparser.add_option_group(group_train)
    group_train.add_option('--device',
                           dest='device',
                           type='string',
                           help='Device for training. We support cpu and cuda now. Auto select if not specified.')
    group_train.add_option('--epochs',
                           dest='epochs',
                           type='int',
                           default=1000,
                           help='Number of maximum epochs for training. Default is 1000.')
    group_train.add_option('--patience',
                           dest='patience',
                           type='int',
                           default=100,
                           help='Number of epochs wait for better result. Default is 100.')
    group_train.add_option('--min-delta',
                           dest='min_delta',
                           type='float',
                           default=0.001,
                           help='Minimum delta for better result. Should be in (0, 1). Default is 0.001.')
    group_train.add_option('--min-epochs',
                           dest='min_epochs',
                           type='int',
                           default=50,
                           help='Minimum number of epochs for training. Default is 50. Set to 0 to disable.')
    group_train.add_option('--batch-size',
                           dest='batch_size',
                           type='int',
                           default=0,
                           help='Batch size for training. Default is 0 for whole dataset.')
    group_train.add_option('-s', '--seed', dest='seed', type='int', help='Random seed for training. Default is random.')
    group_train.add_option('--lr',
                           dest='lr',
                           type='float',
                           default=0.03,
                           help='Learning rate for training. Default is 0.03.')
    return group_train


def add_GCN_options_group(group_train: OptionGroup) -> None:
    """
    Add GCN options group to optparser.
    :param optparser: OptionParser object.
    :return: None.
    """

    # GCN options group
    group_train.add_option('--hidden-feats',
                           dest='hidden_feats',
                           type='int',
                           default=4,
                           help='Number of hidden features. Default is 4.')
    group_train.add_option('--n-gcn-layers',
                           dest='n_gcn_layers',
                           type='int',
                           default=2,
                           help='Number of GCN layers. Default is 2.')


def add_GP_options_group(group_train: OptionGroup) -> None:
    """
    Add Graph Pooling options group to optparser.
    :param optparser: OptionParser object.
    :return: None.
    """

    # NP options group
    group_train.add_option('-k',
                           '--k-clusters',
                           dest='k',
                           type='int',
                           default=6,
                           help='Number of niche clusters. Default is 6. No more than 8.')
    group_train.add_option('--modularity-loss-weight',
                           dest='modularity_loss_weight',
                           type='float',
                           default=0.3,
                           help='Weight for modularity loss. Default is 0.3.')
    group_train.add_option('--purity-loss-weight',
                           dest='purity_loss_weight',
                           type='float',
                           default=300,
                           help='Weight for purity loss. Default is 300.')
    group_train.add_option('--regularization-loss-weight',
                           dest='regularization_loss_weight',
                           type='float',
                           default=0.1,
                           help='Weight for regularization loss. Default is 0.1.')
    group_train.add_option('--beta',
                           dest='beta',
                           type='float',
                           default=0.03,
                           help='Beta value control niche cluster assignment matrix. Default is 0.03.')


def validate_train_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate train options.
    
    Parameters
    ----------
    options : Values
        Options object.
    optparser : Optional[OptionParser], optional
        OptionParser object. The default is None.

    Returns
    -------
    None
    """

    # device
    if getattr(options, 'device', None) is None:
        info('Device not specified, choose automatically.')
    elif options.device.startswith(('cuda', 'cpu')):
        if options.device.startswith('cuda') and not torch.cuda.is_available():
            warning('CUDA is not available, use CPU instead.')
            options.device = 'cpu'
    else:
        warning(f'Invalid device {options.device}! Choose automatically.')
        options.device = None
    if getattr(options, 'device', None) is None:
        if torch.cuda.is_available():
            options.device = 'cuda'
        # elif torch.backends.mps.is_available():  # TODO: torch_geometric do not support MPS now
        #     options.device = 'mps'
        else:
            options.device = 'cpu'

    # epochs
    if getattr(options, 'epochs', None) is None:
        info('epochs is not set. Using default value 1000.')
        options.epochs = 1000
    elif not isinstance(options.epochs, int):
        error(f'epochs must be an integer, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.epochs < 1:
        error(f'epochs must be greater than 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # patience
    if getattr(options, 'patience', None) is None:
        info('patience is not set. Using default value 100.')
        options.patience = 100
    elif not isinstance(options.patience, int):
        error(f'patience must be an integer, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.patience < 1:
        error(f'patience must be greater than 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    
    # min_delta
    if getattr(options, 'min_delta', None) is None:
        info('min_delta is not set. Using default value 0.001.')
        options.min_delta = 0.001
    elif not isinstance(options.min_delta, float):
        error(f'min_delta must be a float, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif not 0 < options.min_delta < 1:
        error(f'min_delta must be in (0, 1), exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # min_epochs
    if getattr(options, 'min_epochs', None) is None:
        info('min_epochs is not set. Using default value 50.')
        options.min_epochs = 50
    elif not isinstance(options.min_epochs, int):
        error(f'min_epochs must be an integer, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.min_epochs < 0:
        error(f'min_epochs must be greater than or equal to 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # batch_size
    if getattr(options, 'batch_size', None) is None:
        info('batch_size is not set. Using default value 0.')
        options.batch_size = 0
    elif not isinstance(options.batch_size, int):
        error(f'batch_size must be an integer, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.batch_size < 0:
        error(f'batch_size must be greater than or equal to 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # seed
    if getattr(options, 'seed', None) is None:
        options.seed = randint(0, 10000)
    
    # lr
    if getattr(options, 'lr', None) is None:
        info('lr is not set. Using default value 0.03.')
        options.lr = 0.03
    elif not isinstance(options.lr, float):
        error(f'lr must be a float, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.lr <= 0:
        error(f'lr must be greater than 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)


def validate_GCN_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate GCN options.
    
    Parameters
    ----------
    options : Values
        Options object.
    optparser : Optional[OptionParser], optional
        Option

    Returns
    -------
    None
    """

    # check hidden_feats
    if getattr(options, 'hidden_feats', None) is None:
        info('hidden_feats is not set. Using default value 4.')
        options.hidden_feats = 4
    elif getattr(options, 'hidden_feats') < 2:
        error(f'hidden_feats must be greater than 1, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    
    # check n_gcn_layers
    if getattr(options, 'n_gcn_layers', None) is None:
        info('n_gcn_layers is not set. Using default value 2.')
        options.n_gcn_layers = 2
    elif getattr(options, 'n_gcn_layers') < 1:
        error(f'n_gcn_layers must be greater than 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)


def validate_GP_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate Graph Pooling options.
    
    
    Parameters
    ----------
    options : Values
        Options object.
    optparser : Optional[OptionParser], optional
        OptionParser object. The default is None.

    Returns
    -------
    None
    """

    # check k
    if getattr(options, 'k', None) is None:
        info('k is not set. Using default value 6.')
        options.k = 6
    elif not isinstance(options.k, int):
        error(f'k must be an integer, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif getattr(options, 'k') < 2:
        error(f'k must be greater than 1, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif getattr(options, 'k') > 8:
        warning(f'We recommend k to be no more than 8. You can set k to a larger value, but it may cause wired results and extra time cost in niche trajectory construction.')
        if getattr(options, 'k') > 10:
            error(f'k must be no more than 10, exit!')
            if optparser is not None: optparser.print_help()
            sys.exit(1)

    # check modularity_loss_weight
    if getattr(options, 'modularity_loss_weight', None) is None:
        info('modularity_loss_weight is not set. Using default value 0.3.')
        options.modularity_loss_weight = 0.3
    elif not isinstance(options.modularity_loss_weight, (float, int)):
        error(f'modularity_loss_weight must be a number, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.modularity_loss_weight < 0:
        error(f'modularity_loss_weight must be greater than or equal to 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # check purity_loss_weight
    if getattr(options, 'purity_loss_weight', None) is None:
        info('purity_loss_weight is not set. Using default value 300.')
        options.purity_loss_weight = 300
    elif not isinstance(options.purity_loss_weight, (float, int)):
        error(f'purity_loss_weight must be a number, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.purity_loss_weight < 0:
        error(f'purity_loss_weight must be greater than or equal to 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # check regularization_loss_weight
    if getattr(options, 'regularization_loss_weight', None) is None:
        info('regularization_loss_weight is not set. Using default value 0.1.')
        options.regularization_loss_weight = 0.1
    elif not isinstance(options.regularization_loss_weight, (float, int)):
        error(f'regularization_loss_weight must be a number, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.regularization_loss_weight < 0:
        error(f'regularization_loss_weight must be greater than or equal to 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # check beta
    if getattr(options, 'beta', None) is None:
        info('beta is not set. Using default value 0.03.')
        options.beta = 0.03
    elif not isinstance(options.beta, (float, int)):
        error(f'beta must be a number, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.beta < 0:
        error(f'beta must be greater than or equal to 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)


def write_train_options_memo(options: Values) -> None:
    """
    Write train options memo to stdout.
    :param options: Options object.
    :return: None.
    """

    info('           -------- train options -------            ')
    info(f'device:  {options.device}')
    info(f'epochs:  {options.epochs}')
    info(f'batch_size:  {options.batch_size}')
    info(f'patience:  {options.patience}')
    info(f'min_delta:  {options.min_delta}')
    info(f'min_epochs:  {options.min_epochs}')
    info(f'seed:  {options.seed}')
    info(f'lr:  {options.lr}')


def write_GCN_options_memo(options: Values) -> None:
    """
    Write GCN options memo to stdout.
    :param options: Options object.
    :return: None.
    """

    info(f'hidden_feats:  {options.hidden_feats}')
    info(f'n_gcn_layers:  {options.n_gcn_layers}')


def write_GP_options_memo(options: Values) -> None:
    """
    Write Graph Pooling options memo to stdout.
    :param options: Options object.
    :return: None.
    """

    info(f'k:  {options.k}')
    info(f'modularity_loss_weight:  {options.modularity_loss_weight}')
    info(f'purity_loss_weight:  {options.purity_loss_weight}')
    info(f'regularization_loss_weight:  {options.regularization_loss_weight}')
    info(f'beta:  {options.beta}')


__all__ = [
    'add_train_options_group',
    'add_GCN_options_group',
    'add_GP_options_group',
    'validate_train_options',
    'validate_GCN_options',
    'validate_GP_options',
    'write_train_options_memo',
    'write_GCN_options_memo',
    'write_GP_options_memo',
]



import os
from pathlib import Path
from typing import Dict, Tuple, Union

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from torch_geometric.loader import DenseDataLoader

from ..data import SpatailOmicsDataset, load_dataset
from ..log import error, info
from ..utils import get_meta_data_file
from .data import load_meta_data, save_cell_type_code


def load_input_data(
    meta_input: Union[str, Path],
    NN_dir: Union[str, Path],
) -> Dict[str, pd.DataFrame]:
    """
    Load data from original inputs.
    :param meta_input: str or Path, meta data file.
    :param NN_dir: str or Path, save directory.
    :return: Dict[str, pd.DataFrame], loaded data.
    """

    output = {}

    # load meta_data
    meta_data_df = load_meta_data(save_dir=NN_dir, meta_data_file=meta_input)
    output['meta_data'] = meta_data_df

    # cell-level data, validate the meta_data
    if 'Cell_Type' not in meta_data_df.columns:
        raise ValueError('Cell_Type in metadata is required for cell-level data.')
    if 'Cell_ID' not in meta_data_df.columns:
        raise ValueError('Cell_ID in metadata is required for cell-level data.')

    return output


def preprocessing_nn(meta_input: Union[str, Path], NN_dir: Union[str, Path]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Preprocessing for niche network.
    :param meta_input: str or Path, meta data file.
    :param NN_dir: str or Path, save directory.
    :return: Tuple[pd.DataFrame, pd.DataFrame], meta data and cell type coding matrix.
    """

    input_data = load_input_data(meta_input=meta_input, NN_dir=NN_dir)

    if 'meta_data' not in input_data:
        raise ValueError('meta_data is required for preprocessing.')
    meta_data_df = input_data['meta_data']

    # generate cell type coding matrix
    meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype('category')
    ct_coding_matrix = np.zeros(shape=(meta_data_df.shape[0],
                                       meta_data_df['Cell_Type'].cat.categories.shape[0]))  # N x #cell_type
    ct_coding_matrix[np.arange(meta_data_df.shape[0]), meta_data_df.Cell_Type.cat.codes.values] = 1
    ct_coding = pd.DataFrame(data=ct_coding_matrix,
                             columns=meta_data_df['Cell_Type'].cat.categories,
                             index=meta_data_df.index)

    # save cell type code
    save_cell_type_code(save_dir=NN_dir, cell_types=meta_data_df['Cell_Type'])

    return meta_data_df, ct_coding


def load_data(NN_dir: Union[str, Path], batch_size: int = 0) -> Tuple[SpatailOmicsDataset, DenseDataLoader]:
    """
    Load data and create sample loader.
    :param NN_dir: str or Path, save directory.
    :param batch_size: int, batch size.
    :return: Tuple[SpatailOmicsDataset, DenseDataLoader], dataset and sample loader.
    """

    info('Loading dataset.')

    dataset = load_dataset(NN_dir=NN_dir)
    batch_size = batch_size if batch_size > 0 else len(dataset)
    sample_loader = DenseDataLoader(dataset, batch_size=batch_size)

    return dataset, sample_loader


def preprocessing_gnn(NN_dir: Union[str, Path],
                      batch_size: int = 0) -> Tuple[SpatailOmicsDataset, DenseDataLoader, pd.DataFrame]:
    """
    Preprocessing for GNN.
    :param NN_dir: str or Path, save directory.
    :param batch_size: int, batch size.
    :return: Tuple[SpatailOmicsDataset, DenseDataLoader, pd.DataFrame], dataset, sample loader, and meta data.
    """

    # meta data
    meta_data_df = pd.read_csv(get_meta_data_file(NN_dir), header=0)
    meta_data_df['Sample'] = meta_data_df['Sample'].astype(str).astype('category')
    meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype(str).astype('category')

    # dataset and sample loader
    dataset, sample_loader = load_data(NN_dir=NN_dir, batch_size=batch_size)

    return dataset, sample_loader, meta_data_df


def preprocessing_nt(NN_dir: Union[str, Path], GNN_dir: Union[str, Path]) -> Tuple[DataFrame, DataFrame, ndarray]:
    """
    Preprocessing for niche trajectory.
    :param NN_dir: str or Path, save directory.
    :param GNN_dir: str or Path, save directory.
    :return: Tuple[DataFrame, DataFrame, ndarray], meta data, niche-level niche cluster assign, and consolidate out_adj.
    """

    # params
    niche_level_niche_cluster_file = Path(f'{GNN_dir}/niche_level_niche_cluster.csv.gz')
    consolidate_out_adj_file = Path(f'{GNN_dir}/consolidate_out_adj.csv.gz')

    # load meta data
    meta_data_df = pd.read_csv(get_meta_data_file(NN_dir), header=0)
    meta_data_df['Sample'] = meta_data_df['Sample'].astype(str).astype('category')
    meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype(str).astype('category')

    # load niche-level niche cluster assign
    if not os.path.exists(niche_level_niche_cluster_file) or not os.path.exists(consolidate_out_adj_file):
        error(f'niche_level_niche_cluster.csv.gz or consolidate_out_adj.csv.gz does not exist in {GNN_dir} directory.')

    niche_level_niche_cluster_assign_df = pd.read_csv(niche_level_niche_cluster_file, header=0, index_col=0)

    # load consolidate out_adj
    consolidate_out_adj_array = np.loadtxt(consolidate_out_adj_file, delimiter=',')

    return meta_data_df, niche_level_niche_cluster_assign_df, consolidate_out_adj_array



from ONTraC.constants import IO_OPTIONS
from ONTraC.optparser._IO import IOOptionsCollection, io_dicts_to_io_options_collection


def test_ontrac_ioc() -> None:
    """
    Test the IO options collection for ONTraC.
    """

    ioc = io_dicts_to_io_options_collection(IO_OPTIONS['ONTraC'])
    assert isinstance(ioc, IOOptionsCollection)
    assert ioc.has_io_option('input')
    assert ioc.get_io_option_attr('input') == 'required'
    assert ioc.has_io_option('NN_dir')
    assert ioc.get_io_option_attr('NN_dir') == 'overwrite'
    assert ioc.has_io_option('GNN_dir')
    assert ioc.get_io_option_attr('GNN_dir') == 'overwrite'
    assert ioc.has_io_option('NT_dir')
    assert ioc.get_io_option_attr('NT_dir') == 'overwrite'
    # wrong options
    assert not ioc.has_io_option('output')
    assert not ioc.has_io_option('log')


def test_ontrac_nn_ioc() -> None:
    """
    Test the IO options collection for ONTraC_NN.
    """

    ioc = io_dicts_to_io_options_collection(IO_OPTIONS['ONTraC_NN'])
    assert isinstance(ioc, IOOptionsCollection)
    assert ioc.has_io_option('input')
    assert ioc.get_io_option_attr('input') == 'required'
    assert ioc.has_io_option('NN_dir')
    assert ioc.get_io_option_attr('NN_dir') == 'overwrite'
    # wrong options
    assert not ioc.has_io_option('GNN_dir')
    assert not ioc.has_io_option('NT_dir')
    assert not ioc.has_io_option('output')
    assert not ioc.has_io_option('log')


def test_ontrac_gnn_ioc() -> None:
    """
    Test the IO options collection for ONTraC_GNN.
    """

    ioc = io_dicts_to_io_options_collection(IO_OPTIONS['ONTraC_GNN'])
    assert isinstance(ioc, IOOptionsCollection)
    assert ioc.has_io_option('NN_dir')
    assert ioc.get_io_option_attr('NN_dir') == 'required'
    assert ioc.has_io_option('GNN_dir')
    assert ioc.get_io_option_attr('GNN_dir') == 'overwrite'
    # wrong options
    assert not ioc.has_io_option('input')
    assert not ioc.has_io_option('NT_dir')
    assert not ioc.has_io_option('output')
    assert not ioc.has_io_option('log')


def test_ontrac_nt_ioc() -> None:
    """
    Test the IO options collection for ONTraC_NT.
    """

    ioc = io_dicts_to_io_options_collection(IO_OPTIONS['ONTraC_NT'])
    assert isinstance(ioc, IOOptionsCollection)
    assert ioc.has_io_option('NN_dir')
    assert ioc.get_io_option_attr('NN_dir') == 'required'
    assert ioc.has_io_option('GNN_dir')
    assert ioc.get_io_option_attr('GNN_dir') == 'required'
    assert ioc.has_io_option('NT_dir')
    assert ioc.get_io_option_attr('NT_dir') == 'overwrite'
    # wrong options
    assert not ioc.has_io_option('input')
    assert not ioc.has_io_option('output')
    assert not ioc.has_io_option('log')


def test_ontrac_gt_ioc() -> None:
    """
    Test the IO options collection for ONTraC_GT.
    """

    ioc = io_dicts_to_io_options_collection(IO_OPTIONS['ONTraC_GT'])
    assert isinstance(ioc, IOOptionsCollection)
    assert ioc.has_io_option('NN_dir')
    assert ioc.get_io_option_attr('NN_dir') == 'required'
    assert ioc.has_io_option('GNN_dir')
    assert ioc.get_io_option_attr('GNN_dir') == 'overwrite'
    assert ioc.has_io_option('NT_dir')
    assert ioc.get_io_option_attr('NT_dir') == 'overwrite'
    # wrong options
    assert not ioc.has_io_option('input')
    assert not ioc.has_io_option('output')
    assert not ioc.has_io_option('log')


def test_ontrac_analysis_ioc() -> None:
    """
    Test the IO options collection for ONTraC_analysis.
    """

    ioc = io_dicts_to_io_options_collection(IO_OPTIONS['ONTraC_analysis'])
    assert isinstance(ioc, IOOptionsCollection)
    assert ioc.has_io_option('NN_dir')
    assert ioc.get_io_option_attr('NN_dir') == 'required'
    assert ioc.has_io_option('GNN_dir')
    assert ioc.get_io_option_attr('GNN_dir') == 'optional'
    assert ioc.has_io_option('NT_dir')
    assert ioc.get_io_option_attr('NT_dir') == 'optional'
    assert ioc.has_io_option('input')
    assert ioc.get_io_option_attr('input') == 'deprecated'
    assert ioc.has_io_option('output')
    assert ioc.get_io_option_attr('output') == 'optional-overwrite'
    assert ioc.has_io_option('log')
    assert ioc.get_io_option_attr('log') == 'optional'


def test_io_options():

    test_ontrac_ioc()
    test_ontrac_nn_ioc()
    test_ontrac_gnn_ioc()
    test_ontrac_nt_ioc()
    test_ontrac_gt_ioc()
    test_ontrac_analysis_ioc()
